from .dios import (
    DictOfSeries,
    _is_dios_like,
    _is_bool_series,
    _is_list_like_not_nested,
    _is_iterator)

import pandas as pd
import pandas.core.common as ccom
import pandas.core.dtypes.common as dcom
_is_list_like = dcom.is_list_like
_is_nested_list_like = dcom.is_nested_list_like
_is_scalar = dcom.is_scalar
_is_integer = dcom.is_integer
_is_dict_like = dcom.is_dict_like
_is_number = dcom.is_number
_is_hashable = dcom.is_hashable
_is_bool_indexer = ccom.is_bool_indexer


class _Indexer:
    def __init__(self, _dios):
        self._dios = _dios
        self._data = _dios._data

    def _unpack_key(self, key):

        key = list(key) if _is_iterator(key) else key

        if isinstance(key, tuple):
            if len(key) > 2:
                raise KeyError("To many indexers")
            rowkey, colkey = key
        else:
            rowkey, colkey = key, slice(None)

        if isinstance(rowkey, tuple) or isinstance(colkey, tuple):
            raise KeyError(f"{key}. tuples are not allowed.")

        rowkey = list(rowkey) if _is_iterator(rowkey) else rowkey
        colkey = list(colkey) if _is_iterator(colkey) else colkey
        return rowkey, colkey


# #############################################################################


class _LocIndexer(_Indexer):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def __getitem__(self, key):

        rowkey, colkey = self._unpack_key(key)
        if _is_dios_like(rowkey) or _is_dios_like(colkey):
            raise ValueError("Cannot index with multidimensional key")

        data = self._data.loc[colkey].copy()

        # .loc[any, scalar] -> (a single) series
        # .loc[scalar, scalar] -> (a single) value
        if _is_hashable(colkey):
            new = data.loc[rowkey]

        # .loc[any, non-scalar]
        else:
            k = "?"
            try:

                for k in data.index:
                    data.at[k] = data.at[k].loc[rowkey]

            except Exception as e:
                raise type(e)(f"failed for column {k}: " + str(e)) from e

            # .loc[scalar, non-scalar] -> column-indexed series
            if _is_hashable(rowkey):
                new = data

            # .loc[non-scalar, non-scalar] -> dios
            else:
                new = self._dios.copy_empty(columns=False)
                new._data = data

        return new

    def __setitem__(self, key, value):

        rowkey, colkey = self._unpack_key(key)
        if _is_dios_like(rowkey) or _is_dios_like(colkey):
            raise ValueError("Cannot index with multidimensional key")

        # .loc[any, scalar]
        if _is_hashable(colkey):
            # .loc[dont-care, new-scalar] = val
            if colkey not in self._dios.columns:
                self._dios._insert(colkey, value)
            else:
                self._data.at[colkey].loc[rowkey] = value

        # .loc[any, non-scalar]
        else:
            i = None
            try:

                for i, s in enumerate(self._data.loc[colkey]):
                    s.loc[rowkey] = value

            except Exception as e:
                raise type(e)(f"failed for "
                              f"column {self._data.loc[colkey].index[i] if i is not None else '?'}: "
                              f"" + str(e)) from e


# #############################################################################


class _iLocIndexer(_Indexer):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def __getitem__(self, key):
        rowkey, colkey = self._unpack_key(key)
        if _is_dios_like(rowkey) or _is_dios_like(colkey):
            raise ValueError("Cannot index with multidimensional key")

        data = self._data.iloc[colkey].copy()

        # .iloc[any, int] -> single series
        # .iloc[int, int] -> single value
        if _is_integer(colkey):
            new = data.iloc[rowkey]

        # .iloc[any, non-int]
        else:
            k = '?'
            try:

                for k in data.index:
                    data.at[k] = data.at[k].iloc[rowkey]

            except Exception as e:
                raise type(e)(f"failed for column {k}: " + str(e)) from e

            # .iloc[int, non-int] -> column-indexed series
            if _is_integer(rowkey):
                new = data

            # .iloc[non-int, non-int] -> dios
            else:
                new = self._dios.copy_empty(columns=False)
                new._data = data

        return new

    def __setitem__(self, key, value):
        rowkey, colkey = self._unpack_key(key)
        if _is_dios_like(rowkey) or _is_dios_like(colkey):
            raise ValueError("Cannot index with multidimensional key")

        # .iloc[any, int]
        if _is_integer(colkey):
            self._data.iat[colkey].iloc[rowkey] = value

        # .iloc[any, non-int]
        else:
            i = None
            try:

                for i, s in enumerate(self._data.iloc[colkey]):
                    s.iloc[rowkey] = value

            except Exception as e:
                raise type(e)(f"failed for "
                              f"column {self._data.loc[colkey].index[i] if i is not None else '?'}: "
                              f"" + str(e)) from e


# #############################################################################


class _AtIndexer(_Indexer):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def _check_key(self, key):
        if not (isinstance(key, tuple) and len(key) == 2
                and _is_hashable(key[0]) and _is_hashable(key[1])):
            raise KeyError(f"{key}. `.at` takes exactly one scalar row-key "
                           "and one scalar column-key")

    def __getitem__(self, key):
        self._check_key(key)
        return self._data.at[key[1]].at[key[0]]

    def __setitem__(self, key, value):
        self._check_key(key)
        if _is_dios_like(value) or _is_nested_list_like(value):
            raise TypeError(".at[] cannot be used to set multi-dimensional values, use .aloc[] instead.")
        self._data.at[key[1]].at[key[0]] = value


# #############################################################################


class _iAtIndexer(_Indexer):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def _check_key(self, key):
        if not (isinstance(key, tuple) and len(key) == 2
                and _is_integer(key[0]) and _is_integer(key[1])):
            raise KeyError(f"{key} `.iat` takes exactly one integer positional "
                           f"row-key and one integer positional scalar column-key")

    def __getitem__(self, key):
        self._check_key(key)
        return self._data.iat[key[1]].iat[key[0]]

    def __setitem__(self, key, value):
        self._check_key(key)
        if _is_dios_like(value) or _is_nested_list_like(value):
            raise TypeError(".iat[] cannot be used to set multi-dimensional values, use .aloc[] instead.")
        self._data.iat[key[1]].iat[key[0]] = value


# #############################################################################


class _aLocIndexer(_Indexer):
    """ align Indexer

    Automatically align (alignable) indexer on all possible axis,
    and handle indexing with non-existent or missing keys gratefully.

    Also align (alignable) values before setting them with .loc
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def __setitem__(self, key, value):
        rowkeys, colkeys, _ = self._unpack_key_aloc(key)

        c = '?'
        try:
            # full-alignable: dios/df, align rows and columns of value to ourself
            # NOTE: this may shrink columns a third time (1st & 2nd in unpack_key_aloc)
            if _is_dios_like(value):
                colkeys = value.columns.intersection(colkeys)
                for i, c in enumerate(colkeys):
                    l = self._data.at[c]
                    r = value[c]
                    idx = l.loc[rowkeys[i]].index.intersection(r.index)
                    l[idx] = r[idx]

            # row-alignable: given series, align rows of value to every
            # (colkeys selected) series in ourself
            elif isinstance(value, pd.Series):
                r, rindex = value, value.index
                for i, c in enumerate(colkeys):
                    l = self._data.at[c]
                    idx = l.loc[rowkeys[i]].index.intersection(rindex)
                    l[idx] = r[idx]

            # if no align is possible, fallback to .loc
            else:
                for i, c in enumerate(colkeys):
                    self._data.at[c].loc[rowkeys[i]] = value

        except Exception as e:
            raise type(e)(f"failed for column {c}: " + str(e)) from e

    def __getitem__(self, key):
        rowkeys, colkeys, lowdim = self._unpack_key_aloc(key)

        c = '?'
        try:

            if lowdim:
                if colkeys:
                    c = colkeys[0]
                    new = self._data.at[c].loc[rowkeys[0]]
                else:
                    new = pd.Series(index=self._dios.iytpe.min_pdindex)
            else:
                data = pd.Series(dtype='O', index=colkeys)
                for i, c in enumerate(data.index):
                    data.at[c] = self._data.at[c].loc[rowkeys[i]]

                new = DictOfSeries(data=data, itype=self._dios.itype,
                                   cast_policy=self._dios._policy,
                                   fastpath=True)

        except Exception as e:
            raise type(e)(f"failed for column {c}: " + str(e)) from e

        return new

    def _unpack_key_aloc(self, key):
        """
        Return a list of row indexer and a list of existing(!) column labels.
        Both list always have the same length and also could be empty together.
        """

        # if a single column-key is given, we will
        # return a single Series, instead of a dios
        lowdim = False

        # dios / df
        if _is_dios_like(key):
            colkey = self._dios.columns.intersection(key.columns)
            rowkey = [self._data.at[c].index.intersection(key[c].index) for c in colkey]

        else:
            rowkey, colkey = self._unpack_key(key)

            # handle gratefully: scalar
            if _is_hashable(colkey):
                colkey = [colkey] if colkey in self._dios.columns else []
                lowdim = True

            # column-alignable: dios, only align on columns, ignore rows
            elif _is_dios_like(colkey):
                colkey = self._dios.columns.intersection(colkey.columns)

            # column-alignable: list-like, filter only existing columns
            elif _is_list_like_not_nested(colkey) and not _is_bool_indexer(colkey):
                colkey = colkey.values if isinstance(colkey, pd.Series) else colkey
                colkey = self._dios.columns.intersection(colkey)

            # not alignable
            # fall back to .loc (boolean list/series, slice(..), ...
            else:
                colkey = self._data.loc[colkey].index

            if len(colkey) == 0:  # (!) `if not colkey:` fails for pd.Index
                return [], [], lowdim

            #
            # filter row key

            # full-alignable: dios/df, align rows and columns
            # NOTE: this may shrink columns a second time
            if _is_dios_like(rowkey):
                colkey = rowkey.columns.intersection(colkey).to_list()
                rowkey = [self._data.at[c].index.intersection(rowkey[c].index) for c in colkey]

            # row-alignable: pd.Series(), align rows to every series in colkey (columns)
            elif isinstance(rowkey, pd.Series):
                rowkey = [self._data.at[c].index.intersection(rowkey.index) for c in colkey]

            # handle gratefully: scalar, transform to row-slice
            elif _is_hashable(rowkey):
                rowkey = [slice(rowkey, rowkey)] * len(colkey)

            # handle gratefully: list-like, filter only existing rows
            # NOTE: dios.aloc[series.index] is processed here
            elif _is_list_like_not_nested(rowkey) and not _is_bool_indexer(rowkey):
                rowkey = [self._data.at[c].index.intersection(rowkey) for c in colkey]

            # not alignable
            # fallback to .loc (processed by caller) - (eg. slice(..), boolean list-like, ...)
            else:
                rowkey = [rowkey] * len(colkey)

        return rowkey, colkey, lowdim

# #############################################################################

